@@ -4,11 +4,7 @@ set -e -x
 cd bindings/python
 
 # Compile wheels
-if [ -f /opt/python/cp36-cp36m/bin/python ];then
-  /opt/python/cp36-cp36m/bin/python setup.py bdist_wheel $@
-else
-  python3 setup.py bdist_wheel $@
-fi
+python3.7 setup.py bdist_wheel $@
 cd dist
 
 # We can't repair an aarch64 wheel on x64 hosts
@@ -16,4 +12,4 @@ cd dist
 if [[ ! "$*" =~ "aarch64" ]];then
   auditwheel repair *.whl
   mv -f wheelhouse/*.whl .
-fi
\ No newline at end of file
+fi
