@@ -46,8 +46,9 @@ static inline void gen_uc_tracecode(TCGContext *tcg_ctx, int32_t size, int32_t t
         0
     };
 
-    if (puc->hooks_count[type] == 1) {
-        cur = puc->hook[type].head;
+    const int hook_type = type & UC_HOOK_IDX_MASK;
+    if (puc->hooks_count[hook_type] == 1) {
+        cur = puc->hook[hook_type].head;
         
         while (cur) {
             hk = cur->data;
